<?php

require_once 'cpear/PEAR.php';
require_once 'cpear/PEAR5.php';
require_once 'cpear/Net/IMAPProtocol.php';
require_once 'cpear/Net/IMAP.php';
require_once 'cpear/Net/Socket.php';
require_once 'cpear/Mail/mimeDecode.php';

class bm_imapclient {
  const MAX_MAIL_PER_REQUEST = 10;

  private $authenticated;
  private $imap;
    
  /**
  * Конструктор
  * 
  * @param string $host Имя хоста
  * @param int $port Порт
  */
  public function __construct($host, $port=143) {
    $this->imap = new Net_IMAP($host, $port);
    $this->imap->_timeout = 30;
    $this->imap->connect($host, $port);
  }

  /**
  * Получить список почтовых ящиков
  *
  * @return array возвращает массив почтовых ящиков или false
  */
  public function getMailboxes() {
    $mb = $this->imap->getMailboxes();
    if($mb instanceof PEAR_Error) {
      return false;
    }

    $output = array();
    foreach($mb as $b) {
      $messages = $this->imap->getNumberOfMessages($b);
      $recent = $this->imap->getNumberOfRecentMessages($b);
      if(is_numeric($messages) and is_numeric($recent)) {
        $output[] = array('name' => $b, 'messages' => $messages, 'recent' => $recent);
      }
    }

    return $output;
  }

  /**
  * Авторизация пользователя
  *
  * @param string $user Имя пользователя
  * @param string $pass Пароль
  * @throws Exception
  */
  public function login($user, $pass) {
    $res = $this->imap->login($user, $pass, false);
    $this->checkResult($res);

    $this->authenticated = true;
  }

  /**
  * выбор ящика $name - название ящика 
  *
  * @param string $name Название почтового ящика
  * @return boolean true - success, false - error 
  */
  public function selectMailbox($name) {
    $res = $this->imap->selectMailbox($name);
    return $res instanceof PEAR_Error ? false : true;
  }

  /**
  * получить список писем с сервера
  *
  * @param int $start - от
  * @param int $num - Количество сообщений
  * @return mixed массив сообщений или false при ошибке
  */
  public function listMessages($start = 1, $num = self::MAX_MAIL_PER_REQUEST) {
    $end = $start + $num;
    if($end > $this->getMessagesCount()) {
      $end = $this->getMessagesCount();
    }
    if($start > $this->getMessagesCount()) {
      return array();
    }
    $this->requireAuthentication();
    $list = $this->imap->getMessagesList("{$start}:{$end}");

    if($list instanceof PEAR_Error) {
      return false;
    }

    return $list;
  }

  /**
  * получить список ID непросмотренных сообщений
  * 
  * @return mixed массив сообщений или false при ошибке
  */
  public function listUnseenMessages() {
    $res = $this->imap->search("UNSEEN");

    if($res instanceof PEAR_Error) {
      return false;
    }

    return $res === null ? array() : $res;
  }

  /**
  * количество писем в почтовом ящике 
  *
  * @ return mixed количество писем или false при ошибке
  */
  public function getMessagesCount() {
    $res = $this->imap->getNumberOfMessages();
    if($res instanceof PEAR_Error) {
      return false;
    }

    return $res;
  }

  /**
  * проверка авторизации
  * 
  * @return boolean true - авторизован, false - нет
  */
  public function isAuthenticated() {
    return $this->authenticated;
  }

  /**
  * получить сообщение с сервера по ID 
  *
  * @param int $id ID сообщения
  * @return mixed MailMessage или false при ошибке
  */
  public function getMessage($id) {
    $this->requireAuthentication();

    $message = $this->imap->getMessages($id);
    if($message) {
      $message = $message[$id];
    } else {
      return false; 
    }

    $flags =  $this->getMessageFlags($id);
    // функция decode занимает очень много времени
    $struct = Mail_mimeDecode::decode(array('include_bodies' => true, 'decode_bodies' => true, 'decode_headers' => 'UTF-8', 'input' => $message));
    $mail = new MailMessage($struct->headers, $flags);
    MailBuilder::build($struct, $mail);
    return $mail;
  }

  /**
  * флаги сообщения 
  * 
  * @param int $id ID сообщения
  * @param array массив флагов сообщения
  */
  public function getMessageFlags($id) {
    $flags = $this->imap->getFlags($id);
    if(!is_array($flags)) {
      return array();
    }
    $output = array();
    foreach($flags[0] as $flag) {
      $output[] = trim(strtolower(str_replace('\\', "", $flag)));
    }

    return $output;
  }

  /**
  * удалить сообщения с сервера
  * 
  * @return boolean true on success
  * @throws Exception
  */
  public function deleteMessages() {
    $this->requireAuthentication();
    $res = $this->imap->deleteMessages();
    $this->checkResult($res);
    $res = $this->imap->expunge();
    $this->checkResult($res);

    return true;
  }

  /**
  * удалить сообщение c $id на сервере 
  *
  * @param int $id ID сообщения
  * @return boolean true on success
  * @throws Exception
  */
  public function deleteMessage($id) {
    $res = $this->imap->deleteMessages($id);
    $this->checkResult($res);
    $res = $this->imap->expunge();
    $this->checkResult($res);

    return true;
  }

  /**
  * проверка результата запроса
  *
  * @throws new Exception
  */
  private function checkResult($res) {
    if($res instanceof PEAR_Error) {
      throw new Exception($res->message);
    }
  }

  /**
  * Запрос авторизации
  *
  * @throws new Exception
  */
  private function requireAuthentication() {
    if(!$this->authenticated) {
      throw new Exception("not authenticated");
    }
  }

  /**
  * Destructor
  */
  public function __desctruct() {
    $this->imap->disconnect();
  }
}


class MailBuilder {
  /**
  * Собирает MailMessage из частей 
  * 
  * @param string $part исходная часть сообщения
  * @param MailMessage &$mail резултат
  * @param boolean $body определяет часть сообщения(тело или вложение)
  */
  public static function build($part, &$mail, $body=true) {
    switch ($part->ctype_primary) {
      case 'multipart':
        foreach($part->parts as $p) {
          self::build($p, $mail, $body);
          if($part->ctype_secondary !== 'alternative') {
            $body = false;
          }
        }
        break;
      case 'text':
        if($body) {
          $charset = isset($part->ctype_parameters['charset']) ? $part->ctype_parameters['charset'] : false;
          if($charset) {
            $body = @iconv($charset, 'UTF-8', $part->body);
          }
          $mail->setBody($body, $part->ctype_secondary);
          break;
        }
      default:
        $mail->addAttachment(
          new MailAttachment(
            $part->headers, 
            isset($part->body) ? $part->body : ((isset($part->parts) && count($part->parts) > 0) ? $part->parts[0]->body : ''), 
            isset($part->ctype_parameters['name']) ? $part->ctype_parameters['name'] : null, 
            isset($part->d_parameters['filename']) ? $part->d_parameters['filename'] : null,
            isset($part->disposition) ? $part->disposition : null,
            "{$part->ctype_primary}/{$part->ctype_secondary}"
            ));
    }
  }

}

class MailAttachment {
  private $headers = array();
  private $fileName;
  private $body;
  private $name;
  private $disposition;
  private $type;

  /**
  * конструктор
  * 
  * @param array $headers заголовки вложения
  * @param string $body тело вложения
  * @param string $name имя вложения
  * @param string $fileName имя файла
  * @param string $disposition тип вложения (inline/attachment)
  * @param string $type mime/type вложения
  */
  public function __construct($headers, $body, $name, $fileName, $disposition, $type) {
    $this->fileName = trim($fileName);
    $this->headers = $headers;
    $this->body = $body;
    $this->name = trim($name);
    $this->disposition = trim($disposition);
    $this->type = $type;
  }

  /**
  * получить ID прикрепленного файла 
  * 
  * @return mixed ID вложения или false
  */
  public function getId() {
    return isset($this->headers['content-id']) ? str_replace(array('<','>'), '', $this->headers['content-id']) : false;
  }

  /*
  * тип вложения
  * 
  * @return string inline/attachment
  */
  public function getDisposition() {
    return $this->disposition;
  }

  /* 
  * https://en.wikipedia.org/wiki/Data_URI_scheme 
  *
  * @return string data url for attachment
  */
  public function getDataUri() {
    return "data:{$this->type};base64," . base64_encode($this->body);
  }

  /**
  * получить контент вложения
  *
  * @return string контент вложения
  */
  public function getBody() {
    return $this->body;
  }

  /**
  * получить имя файла 
  *
  * @return string attachment name
  */
  public function getFileName() {
    return is_null($this->fileName) || empty($this->fileName)  ? $this->name : $this->fileName;
  }

  /**
  * получить mime type вложения
  * 
  * @return string
  */
  public function getMimeType() {
    return $this->type;
  }

  /* 
  * содержимое заголовка по названию
  * 
  * @param string $key название заголовка
  * @return mixed значение заголовка или false
  */
  public function getHeader($key) {
    return isset($this->headers[$key]) ? $this->headers[$key] : false;
  }

  /**
  * сохранить файл на диске 
  *
  * @param string $dir dirname/filename
  * @return boolean true on success, false on error
  */
  public function save($dir) {
    $dir = rtrim($dir,DIRECTORY_SEPARATOR);

    if(is_dir($dir)) {
      $file = $this->getFileName();
      if(empty($file)) {
        return false;
      }
      $dir = $dir . DIRECTORY_SEPARATOR . $this->getFileName();
    }

    if(file_exists($dir) && !is_writable($dir)) {
        return false;
    }

    if(is_writable(dirname($dir))) {
      return file_put_contents($dir, $this->body);
    }

    return false;
  }

}

class MailMessage {

  private $headers = array();
  private $from;
  private $to;
  private $text = false;
  private $html = false;
  private $formattedHTML;
  private $attachments = array();
  private $cc = array();
  private $flags = array();

  /**
  * Constructor
  *
  * @param array $headers Заголовки сообщения
  * @param array $flags Флаги сообщения
  */
  public function __construct($headers, $flags) {
    $this->headers = $headers;
    $this->flags = $flags;
  }

  /**
  * Возвращает отправителя письма
  *
  * @return MailAddress from mail
  */
  public function getFrom() {
    if(!$this->from) {
      $this->from = MailTools::parseAddress($this->headers['from']);
    }

    return $this->from;
  }

  /**
  * возвращает получателя письма
  * 
  * @return array массив объектов MailAddress
  */
  public function getTo() {
    if(!$this->to) {
      $this->to = MailTools::parseAddresses($this->headers['to']);
    }

    return $this->to;
  }

  /**
  * Возвращает получателей копии письма
  * 
  * @return array массив объектов MailAddress
  */
  public function getCC() {
    $ccHeader = $this->getHeader('cc');
    if($ccHeader && $this->cc) {
      $this->cc = MailTools::parseAddresses($ccHeader);
    }

    return $this->cc;
  }

  /**
  * Возвращает тему письма
  * 
  * @return string тема письма
  */
  public function getSubject() {
    return $this->subject = $this->headers['subject'];
  }

  /**
  * получить содержимое заголовка по названию 
  * 
  * @param $key название заголовка
  * @return string содержимое заголовка или false
  */
  public function getHeader($key) {
    return isset($this->headers[$key]) ? $this->headers[$key] : false;
  }

  /**
  * Возвращает text/plain часть сообщения 
  * 
  * @return string текст сообщения
  */
  public function getText() {
    return $this->text;
  }

  /**
  * Возвращает text/html часть сообщения
  * 
  * @return string html текст сообщения
  */
  public function getHTML() {
    return $this->html;
  }

  /**
  * Возвращает отформатированный html с inline вложениями 
  *
  * @return string отформатированный html
  */
  public function getFormattedHTML() {
    if(!$this->html) {
      return $this->html;
    }

    if(!$this->formattedHTML) {
      $output = $this->html;
      foreach($this->attachments as $attachment) {
        if($attachment->getDisposition() == 'inline' && $attachment->getId()) {
          $output = str_replace("cid:{$attachment->getId()}", $attachment->getDataUri(), $output);
        }
      }

      $this->formattedHTML = $output;
    }

    return $this->formattedHTML;
  }

  /**
  * Установить текст сообщения по типу 
  *
  * @param string $body текст сообщения
  * @param string type html или plain
  */
  public function setBody($body, $type) {
    if($type == 'html') {
      $this->html = $body;
    } else {
      $this->text = $body;
    }
  }

  /**
  * добавить вложение к сообщению
  * @param MailAttachment $attachment вложение
  */
  public function addAttachment($attachment) {

    if($attachment->getId()) {
      $this->attachments[$attachment->getId()] = $attachment;
    } else {
      $this->attachments[] = $attachment;
    }
  }

  /**
  * Возвращает количество вложений в письме по типу($type)
  *
  * @param string $type тип сообщения
  * @return int количество вложений
  */
  public function attachmentsCount($type) {
    return count($this->getAttachments($type));
  }

  /**
  * Возвращает массив вложений по типу $type
  * Возвращает массив всех вложений если $type == ''
  *
  * @param string $type тип вложения (attachment/inline)
  * @return array массив MailAttachment
  */
  public function getAttachments($type = '') {
    if(!$type) {
      return $this->attachments;
    }

    $output = array();
    foreach($this->attachments as $attachment) {
      if($attachment->getDisposition() == $type) {
        $output[] = $attachment;
      }
    }

    return $output;
  }

  /**
  * Возвращает флаги сообщения
  *
  * @return array
  */
  public function getFlags() {
    return $this->flags;
  }

  /**
  * Возвращает дату доставки сообщения 
  *
  * @return int timestamp
  */
  public function getDate() {
    $date = $this->getHeader('date');
    if(!$date) {
      return false;
    }
    return strtotime($date);
  }

  /**
  * Возвращает true если есть флагaг Draft
  * 
  * @return boolean
  */
  public function isDraft() {
    return $this->hasFlag('Draft');
  }

  /**
  * просмотрено? 
  *
  * @return boolean
  */
  public function isSeen() {
    return $this->hasFlag('Seen');
  }

  /**
  * удалено?
  *
  * @return boolean
  */
  public function isDeleted() {
    return $this->hasFlag('Deleted');
  }

  /**
  * Проверяет наличие флага у сообщения
  *
  * @param string $flag
  * @return boolean
  */
  public function hasFlag($flag) {
    $flag = strtolower($flag);
    return in_array($flag, $this->flags);
  }
}

class MailAddress {
  private $name;
  private $address;

  /**
  * @param string $name - Имя владельца адреса
  * @param string $address - адрес
  */
  public function __construct($name, $address) {
    $this->name = $name;
    $this->address = $address;
  }

  /**
  * Возвращает имя владельца адреса(если указано)
  *
  * @return string
  */
  public function getName() {
    return $this->name;
  }

  /**
  * адрес
  *
  * @return string
  */
  public function getAddress() {
    return $this->address;
  }

  /**
  * Magic XD
  *
  * @return string
  */
  public function __toString() {
    return !empty($this->name) ? "{$this->name} <{$this->address}>" : $this->address;
  }
}


class MailTools {

  /**
  * Парсит строку в массив
  * 
  * @param string $str строка с адресом
  * @return array
  */
  public static function parseAddresses($str) {
    $addresses = explode(",", $str);
    if(count($addresses) == 0) {
      return array(self::parseAddress($str));
    }
    $output = array();
    foreach($addresses as $address) {
      $output[] = self::parseAddress($address);
    }

    return $output;
  }

  /**
  * Парсит строку в MailAddress 
  *
  * @param string $address
  * @return MailAddress
  */
  public static function parseAddress($address) {
    if(preg_match("/([^<]+)<([^>]+)>/", $address, $matches)) {
      return new MailAddress(trim($matches[1]), trim($matches[2]));
    }

    return new MailAddress('', trim($address));
  }
}
